@@ -27,10 +27,15 @@
  * MT safe
  */
 
-#include "glib.h"
+#include "config.h"
+
 #include <stdarg.h>
 #include <string.h>
 
+#include "glib.h"
+#include "galias.h"
+
+
 typedef struct _GRealTuples        GRealTuples;
 
 struct _GRelation
@@ -40,7 +45,6 @@ struct _GRelation
   
   GHashTable   *all_tuples;
   GHashTable  **hashed_tuple_tables;
-  GMemChunk    *tuple_chunk;
   
   gint count;
 };
@@ -65,9 +69,19 @@ tuple_equal_2 (gconstpointer v_a,
 static guint
 tuple_hash_2 (gconstpointer v_a)
 {
+#if GLIB_SIZEOF_VOID_P > GLIB_SIZEOF_LONG
+  /* In practise this snippet has been written for 64-bit Windows
+   * where ints are 32 bits, pointers 64 bits. More exotic platforms
+   * need more tweaks.
+   */
+  guint* a = (guint*) v_a;
+
+  return (a[0] ^ a[1] ^ a[2] ^ a[3]);
+#else
   gpointer* a = (gpointer*) v_a;
   
   return (gulong)a[0] ^ (gulong)a[1];
+#endif
 }
 
 static GHashFunc
@@ -104,16 +118,22 @@ g_relation_new (gint fields)
   GRelation* rel = g_new0 (GRelation, 1);
   
   rel->fields = fields;
-  rel->tuple_chunk = g_mem_chunk_new ("Relation Chunk",
-				      fields * sizeof (gpointer),
-				      fields * sizeof (gpointer) * 128,
-				      G_ALLOC_AND_FREE);
   rel->all_tuples = g_hash_table_new (tuple_hash (fields), tuple_equal (fields));
   rel->hashed_tuple_tables = g_new0 (GHashTable*, fields);
   
   return rel;
 }
 
+static void
+relation_delete_value_tuple (gpointer tuple_key,
+                             gpointer tuple_value,
+                             gpointer user_data)
+{
+  GRelation *relation = user_data;
+  gpointer *tuple = tuple_value;
+  g_slice_free1 (relation->fields * sizeof (gpointer), tuple);
+}
+
 static void
 g_relation_free_array (gpointer key, gpointer value, gpointer user_data)
 {
@@ -127,9 +147,6 @@ g_relation_destroy (GRelation *relation)
   
   if (relation)
     {
-      g_hash_table_destroy (relation->all_tuples);
-      g_mem_chunk_destroy (relation->tuple_chunk);
-      
       for (i = 0; i < relation->fields; i += 1)
 	{
 	  if (relation->hashed_tuple_tables[i])
@@ -138,6 +155,9 @@ g_relation_destroy (GRelation *relation)
 	      g_hash_table_destroy (relation->hashed_tuple_tables[i]);
 	    }
 	}
+
+      g_hash_table_foreach (relation->all_tuples, relation_delete_value_tuple, relation);
+      g_hash_table_destroy (relation->all_tuples);
       
       g_free (relation->hashed_tuple_tables);
       g_free (relation);
@@ -161,16 +181,16 @@ void
 g_relation_insert (GRelation   *relation,
 		   ...)
 {
-  gpointer* tuple = g_chunk_new (gpointer, relation->tuple_chunk);
+  gpointer* tuple = g_slice_alloc (relation->fields * sizeof (gpointer));
   va_list args;
   gint i;
   
-  va_start(args, relation);
+  va_start (args, relation);
   
   for (i = 0; i < relation->fields; i += 1)
-    tuple[i] = va_arg(args, gpointer);
+    tuple[i] = va_arg (args, gpointer);
   
-  va_end(args);
+  va_end (args);
   
   g_hash_table_insert (relation->all_tuples, tuple, tuple);
   
@@ -206,21 +226,21 @@ g_relation_delete_tuple (gpointer tuple_key,
 			 gpointer user_data)
 {
   gpointer      *tuple = (gpointer*) tuple_value;
-  GRelation     *rel = (GRelation *) user_data;
+  GRelation     *relation = (GRelation *) user_data;
   gint           j;
   
   g_assert (tuple_key == tuple_value);
   
-  for (j = 0; j < rel->fields; j += 1)
+  for (j = 0; j < relation->fields; j += 1)
     {
-      GHashTable *one_table = rel->hashed_tuple_tables[j];
+      GHashTable *one_table = relation->hashed_tuple_tables[j];
       gpointer    one_key;
       GHashTable *per_key_table;
       
       if (one_table == NULL)
 	continue;
       
-      if (j == rel->current_field)
+      if (j == relation->current_field)
 	/* can't delete from the table we're foreaching in */
 	continue;
       
@@ -231,9 +251,10 @@ g_relation_delete_tuple (gpointer tuple_key,
       g_hash_table_remove (per_key_table, tuple);
     }
   
-  g_hash_table_remove (rel->all_tuples, tuple);
+  if (g_hash_table_remove (relation->all_tuples, tuple))
+    g_slice_free1 (relation->fields * sizeof (gpointer), tuple);
   
-  rel->count -= 1;
+  relation->count -= 1;
 }
 
 gint
@@ -241,11 +262,15 @@ g_relation_delete  (GRelation     *relation,
 		    gconstpointer  key,
 		    gint           field)
 {
-  GHashTable *table = relation->hashed_tuple_tables[field];
+  GHashTable *table; 
   GHashTable *key_table;
-  gint        count = relation->count;
+  gint        count;
   
   g_return_val_if_fail (relation != NULL, 0);
+
+  table = relation->hashed_tuple_tables[field];
+  count = relation->count;
+
   g_return_val_if_fail (table != NULL, 0);
   
   key_table = g_hash_table_lookup (table, key);
@@ -289,14 +314,18 @@ g_relation_select (GRelation     *relation,
 		   gconstpointer  key,
 		   gint           field)
 {
-  GHashTable  *table = relation->hashed_tuple_tables[field];
+  GHashTable  *table;
   GHashTable  *key_table;
-  GRealTuples *tuples = g_new0 (GRealTuples, 1);
+  GRealTuples *tuples; 
   gint count;
   
   g_return_val_if_fail (relation != NULL, NULL);
+
+  table = relation->hashed_tuple_tables[field];
+
   g_return_val_if_fail (table != NULL, NULL);
   
+  tuples = g_new0 (GRealTuples, 1);
   key_table = g_hash_table_lookup (table, key);
   
   if (!key_table)
@@ -319,10 +348,13 @@ g_relation_count (GRelation     *relation,
 		  gconstpointer  key,
 		  gint           field)
 {
-  GHashTable  *table = relation->hashed_tuple_tables[field];
+  GHashTable  *table;
   GHashTable  *key_table;
   
   g_return_val_if_fail (relation != NULL, 0);
+
+  table = relation->hashed_tuple_tables[field];
+
   g_return_val_if_fail (table != NULL, 0);
   
   key_table = g_hash_table_lookup (table, key);
@@ -336,7 +368,7 @@ g_relation_count (GRelation     *relation,
 gboolean
 g_relation_exists (GRelation   *relation, ...)
 {
-  gpointer* tuple = g_chunk_new (gpointer, relation->tuple_chunk);
+  gpointer *tuple = g_slice_alloc (relation->fields * sizeof (gpointer));
   va_list args;
   gint i;
   gboolean result;
@@ -350,7 +382,7 @@ g_relation_exists (GRelation   *relation, ...)
   
   result = g_hash_table_lookup (relation->all_tuples, tuple) != NULL;
   
-  g_mem_chunk_free (relation->tuple_chunk, tuple);
+  g_slice_free1 (relation->fields * sizeof (gpointer), tuple);
   
   return result;
 }
@@ -404,7 +436,7 @@ g_relation_print_one (gpointer tuple_key,
     }
   
   g_string_append (gstring, "]");
-  g_log (g_log_domain_glib, G_LOG_LEVEL_INFO, gstring->str);
+  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s", gstring->str);
   g_string_free (gstring, TRUE);
 }
 
@@ -416,7 +448,7 @@ g_relation_print_index (gpointer tuple_key,
   GRelation* rel = (GRelation*) user_data;
   GHashTable* table = (GHashTable*) tuple_value;
   
-  g_log (g_log_domain_glib, G_LOG_LEVEL_INFO, "*** key %p", tuple_key);
+  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "*** key %p", tuple_key);
   
   g_hash_table_foreach (table,
 			g_relation_print_one,
@@ -428,7 +460,7 @@ g_relation_print (GRelation *relation)
 {
   gint i;
   
-  g_log (g_log_domain_glib, G_LOG_LEVEL_INFO, "*** all tuples (%d)", relation->count);
+  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "*** all tuples (%d)", relation->count);
   
   g_hash_table_foreach (relation->all_tuples,
 			g_relation_print_one,
@@ -439,7 +471,7 @@ g_relation_print (GRelation *relation)
       if (relation->hashed_tuple_tables[i] == NULL)
 	continue;
       
-      g_log (g_log_domain_glib, G_LOG_LEVEL_INFO, "*** index %d", i);
+      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "*** index %d", i);
       
       g_hash_table_foreach (relation->hashed_tuple_tables[i],
 			    g_relation_print_index,
@@ -447,3 +479,6 @@ g_relation_print (GRelation *relation)
     }
   
 }
+
+#define __G_REL_C__
+#include "galiasdef.c"
